/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/

#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <pty.h>
#include <stdio.h>
#include <utmp.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>

#include <viper.h>

#include "vwmterm.h"
#include "watchdog.h"
#include "whiteboard.h"
#include "scrollback.h"
#include "events.h"
#include "psthread.h"
#include "signals.h"

G_MODULE_EXPORT gchar* g_module_check_init(GModule *module)
{
	gchar			*libfilename;

	/*	preload libgtop for use with this module	*/
	if(g_module_open("libutil.so",G_MODULE_BIND_LAZY)==NULL)
 		return "\ncould not preload libutil.so";

	/*	register with vwm 	*/
	libfilename=(gchar*)g_module_name(module);
	vwm_module_add("Tools","Terminal Shell",vwmterm_init,NULL,libfilename);

	return NULL;
}

WINDOW* vwmterm_init(gpointer anything)
{
   extern WINDOW  	*SCREEN_WINDOW;
	extern ps_runq_t	*vwm_runq;
	WINDOW	      	*window;
	vwmterm_t			*vwmterm;
	vwmterm_t			vwmterm_clone;
	struct termios		term_desc;
	gint		      	width,height;
	gchar		      	*termvar;
   gchar          	*title;
	gint		      	retval;
	int					fflags;

   getmaxyx(SCREEN_WINDOW,height,width);
   if(height>30 && width>84)
   {
      height=25;
      width=80;
   }
   else
   {
      /* calculate scaled window size */
	   window_get_size_scaled(NULL,&width,&height,0.85,0.65);
	   if(width>80) width=80;
	   if(height>25) height=25;
   }

	/* create pseudo-terminal */
	vwmterm=(vwmterm_t*)g_malloc0(sizeof(vwmterm_t));
	vwmterm->win_sz.ws_row=height;
	vwmterm->win_sz.ws_col=width;
	tcgetattr(STDIN_FILENO,&term_desc);

	/*	default terminal settings - added 2/6/2006	*/
	/* compatibility update on 5/2/2007					*/
	/*	compatibility update on 3/31/2009				*/
	term_desc.c_lflag |= ECHO;
	term_desc.c_lflag |= ISIG;
	term_desc.c_lflag |= ICANON;
	term_desc.c_lflag |= IEXTEN;

	term_desc.c_iflag |= BRKINT;
	term_desc.c_iflag |= IGNPAR;
	term_desc.c_iflag |= IXON;
 	term_desc.c_iflag |= ICRNL;	/*	this has to be set for su & ssh	*/
	term_desc.c_iflag &= ~IXANY;
	term_desc.c_iflag &= ~IGNCR;  /* must be off for ICRNL to work		*/
	term_desc.c_oflag |= OPOST;	/*	enables ouput post-processing		*/
	term_desc.c_oflag |= TAB3;		/*	convert tabs to spaces				*/
	term_desc.c_oflag |= ONLRET;	/*	NL does peform carriage-return	*/
	term_desc.c_oflag &= ~ONLCR;	/* do not xlate NL to CR-LF			*/
	term_desc.c_oflag &= ~OCRNL;	/* don't xlate CR to NL					*/
	term_desc.c_oflag &= ~ONOCR;	/* allow CR at column 0					*/
	term_desc.c_oflag &= ~OFILL;	/* do not send fill chars as delay	*/
#ifdef OLCUC
	term_desc.c_oflag &= ~OLCUC;	/* don't xlate lower case to upper	*/
#endif

	retval=openpty(&vwmterm->master_fd,&vwmterm->slave_fd,NULL,
		&term_desc,&vwmterm->win_sz);
	if(retval==-1) return NULL;
	sprintf(vwmterm->term_name," Terminal Shell: %s ",
      ttyname(vwmterm->slave_fd));
   fflush(NULL);

   /* copy the contents of vwmterm (which is on the heap) into vwmterm_clone
      which is on the stack.  this ensures that the child process will have
      correct references to file descriptors, etc. because the stack gets
      cloned during a fork(). */
   memcpy(&vwmterm_clone,vwmterm,sizeof(vwmterm_t));
   vwmterm_clone.child_pid=fork();
   if(vwmterm_clone.child_pid==0)
	{
      close(vwmterm_clone.master_fd);
      if(login_tty(vwmterm_clone.slave_fd)==-1) return NULL;
		signal(SIGINT,SIG_DFL);
      dup2(vwmterm_clone.slave_fd,STDERR_FILENO);

		termvar=getenv("SHELL");
		if(termvar==NULL) termvar="/bin/sh";
      title=g_strrstr(termvar,"/");
      if(title!=NULL) title++;
      if(title[0]=='\0') title="sh";
      setenv("TERM","vwmterm",1);
      setenv("COLORTERM","vwmterm",1);
      setenv("LS_OPTIONS"," --color=always",1);
      execl(termvar,title,"-i",(char*)NULL);
      close(vwmterm_clone.slave_fd);
		exit(0);
	}

   vwmterm_sigset(SIGIO,vwmterm_SIGIO);
	fcntl(vwmterm->master_fd,F_SETOWN,getpid());
   fflags=fcntl(vwmterm->master_fd,F_GETFL);
   fcntl(vwmterm->master_fd,F_SETFL,fflags | FASYNC);

   /* we used pty_clone for the fork so we need to make sure that pty_term
      has the correct child PID reference.   */
   vwmterm->child_pid=vwmterm_clone.child_pid;

   /* create window  */
   viper_thread_enter();
	window=viper_window_create(vwmterm->term_name,0.5,0.5,width,height,TRUE);
   viper_window_set_state(window,STATE_UNSET | STATE_NORESIZE);
	viper_window_set_limits(window,15,2,WSIZE_UNCHANGED,WSIZE_UNCHANGED);

	vwmterm->vscreens[SCROLLBACK]=VSCREEN(scrollback_vscreen_from(window,
		vwmterm->master_fd));
	vwmterm->vscreens[WHITEBOARD]=whiteboard_vscreen_from(window,
		vwmterm->master_fd);

	viper_thread_leave();

   /* initialize these so that VWMTERM_IS_FG and VWMTERM_WAS_FG macros
      will both return true.  */
	vwmterm->window=window;
	vwmterm->new_state |= PTY_FG;
	vwmterm->old_state |= PTY_FG;

	psthread_add(vwm_runq,vwmterm_psthread,(gpointer)vwmterm);

	viper_thread_enter();
	viper_event_set(window,"window-resized",vwmterm_ON_RESIZE,NULL);
	viper_event_set(window,"window-close",vwmterm_ON_CLOSE,(gpointer)vwmterm);
	viper_event_set(window,"window-destroy",vwmterm_ON_DESTROY,
		(gpointer)vwmterm);
	viper_window_set_key_func(window,vwmterm_ON_KEYSTROKE);
	viper_window_set_userptr(window,(gpointer)vwmterm);
	viper_thread_leave();

	return window;
}


