/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <pty.h>
#include <utmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <wait.h>
#include <poll.h>
#include <ctype.h>
#include <errno.h>

#include <viper.h>

#include "psthread.h"
#include "vwmterm.h"
#include "vscreen.h"
#include "watchdog.h"
#include "xlate.h"
#include "events.h"


gint vwmterm_psthread(ps_task_t *task,gpointer anything)
{
	static gchar	*acs="`afgjklmnopqrstuvwxyz{|}~";
	WINDOW  			*window;
	vwmterm_t		*vwmterm;
	vscreen_t		*vscreen=NULL;				/*	convenience ptr	*/
	gint				retval=0;
	pid_t				child_pid=0;
	gchar				*buffer=NULL;
   gchar          *carryover;
	gchar				*pos_ptr;
   gint           bytes_waiting=0;
	gint				bytes_to_read=0;
	gint			   bytes_read=0;
	gint				bytes_xlate=0;
	gint				x,y;
	struct pollfd	data_poll;
	chtype			cell;
	gchar				*token;
#ifdef _DEBUG
	int				debug_fd;
	gchar				*filename;
#endif


	vwmterm=(vwmterm_t*)anything;
	window=(WINDOW*)vwmterm->window;

	/*	if child_pid==pty_term->child_pid then it is very likely that the shell
		(ie. bash) terminated normally.  in this case we can skip the
      "window-close" event and proceed to window destruction.  alternately,
      when when WHNOHANG is set, -1 indicates an error.  when this happens, it
		probably means that the user closed the window using the VWM
		window close button and the event "window-close" has already been
   	handled. */
	child_pid=waitpid(vwmterm->child_pid,NULL,WNOHANG);
	if(child_pid==vwmterm->child_pid || child_pid==-1)
	{
		viper_thread_enter();
		viper_window_destroy(window);
		viper_thread_leave();
		return PSTHREAD_TERMINATE;
	}

   data_poll.fd=vwmterm->master_fd;
	data_poll.events=POLLIN;
   retval=poll(&data_poll,1,10);

	if(retval>0)
	{
 		/*	find out how much data is availbe for reading and allocate
			a buffer sufficent to store the data.  one of these _should_
         be defined on most *nix systtems.   */
#ifdef FIONREAD
		ioctl(vwmterm->master_fd,FIONREAD,&bytes_waiting);
#else
      ioctl(vwmterm->master_fd,TIOCOUTQ,&bytes_waiting);
#endif

      bytes_to_read=bytes_waiting;
      if(bytes_to_read>PTY_READ_MAX) bytes_to_read=PTY_READ_MAX;
 		buffer=(gchar*)g_malloc0(bytes_to_read+10); /* 10 byte padding */
		pos_ptr=buffer;
		bytes_read=read(vwmterm->master_fd,buffer,bytes_to_read);
		if(bytes_read==0)
      {
         g_free(buffer);
         return PSTHREAD_CONTINUE;
      }

      /* if read() returns -1, something bad has happened.  most likely the
         shell died (for some reason, maybe user sent SIGKILL).  if this
         happens we need to clean up gracefully.   */
      if(bytes_read==-1)
      {
         g_free(buffer);
         viper_thread_enter();
         viper_window_close(window);
         viper_thread_leave();
         /* allow this pseudo thread to run one more time so that waitpid()
            will catch the exit on next iteration.   */
         return PSTHREAD_CONTINUE;
      }

      /* prepend any broken escape sequences which might have been carried
         over from previous read */
      if(vwmterm->carryover!=NULL)
      {
         carryover=(gchar*)g_malloc0(vwmterm->carry_sz+bytes_to_read+10);
         memcpy(carryover,vwmterm->carryover,vwmterm->carry_sz);
         memcpy(&carryover[vwmterm->carry_sz],buffer,bytes_to_read);
         g_free(buffer);
         buffer=carryover;
         g_free(vwmterm->carryover);
         vwmterm->carryover=NULL;
         bytes_read=bytes_to_read+vwmterm->carry_sz;
         pos_ptr=buffer;
      }

#ifdef _DEBUG
		filename=g_strdup_printf("log-vwmterm-%d.out",vwmterm->child_pid);
		debug_fd=open(filename,O_CREAT | O_WRONLY | O_APPEND);
		g_free(filename);
		write(debug_fd,(void*)pos_ptr,(size_t)bytes_read);
		close(debug_fd);
#endif

		viper_thread_enter();
		while(bytes_read>0)
		{
			/*	this has to be updated each iteration because some of the
				xlate functions can change the active vscreen.	*/
			vscreen=VWMTERM_SCREEN(vwmterm);

         /* if bytes_to_read==PTY_READ_MAX, it almost certainly means ouput
            from the slave process was broken up in chunks.  this can cause
            broken escape sequences which won't get parsed correctly.  they
            need to be treated as carryover for the next read.  the saftey
            margin of 20 bytes is arbitrarily chosen.  i've never seen an
            escape code sequence longer than 20 bytes--but it could happen. */
         if((bytes_waiting>bytes_to_read) && (bytes_read<20))
         {
            vwmterm_carryover(vwmterm,pos_ptr,bytes_read);
            break;
         }

			/*	discard useless escape codes	*/
			bytes_xlate=vwmterm_xlate_discards(pos_ptr);
			if(bytes_xlate>0)
			{
				bytes_read-=bytes_xlate;
				pos_ptr+=bytes_xlate;
				continue;
			}

			/* these are special 1-byte control codes.   */
			bytes_xlate=vwmterm_xlate_specials(vwmterm,pos_ptr);
			if(bytes_xlate>0)
			{
				pos_ptr+=bytes_xlate;
 				bytes_read--;
				continue;
			}
         if(bytes_xlate==-1)
         {
            vwmterm_carryover(vwmterm,pos_ptr,bytes_read);
            break;
         }

			bytes_xlate=vwmterm_xlate_esc(vwmterm,pos_ptr);
         /* a return value of -1 indicates a broken escape sequence.  this
            is a timing problem which occurs with the read/write master/slave
            pty buffers fill up.  the solution is to store the incomplete
            sequence and prepend on the next read. */
         if(bytes_xlate==-1)
         {
            vwmterm_carryover(vwmterm,pos_ptr,bytes_read);
            break;
         }
         else vwmterm->carryover=NULL;

			if(bytes_xlate==0)
			{
				cell=pos_ptr[0];
				if(vscreen->state & VSCREEN_ACS)
				{
					token=strchr(acs,pos_ptr[0]);
					if(token!=NULL) cell=NCURSES_ACS(pos_ptr[0]);
				}

				vscreen_addch(vscreen,cell);
				pos_ptr++;
				bytes_read--;
			}
			else
			{
				pos_ptr+=bytes_xlate;
				bytes_read-=bytes_xlate;
			}
		}

		if(buffer!=NULL) g_free(buffer);
		getmaxyx(window,y,x);
		/* it's not well documented, but getmaxyx() does not actually return the
			maximum value for x and y coords, but rather the width and height of
			a window.  the former is zero indexed while the latter is index+1.
			for this reason it is necessary to subtract 1 from y and x.	*/
		copywin(VWMTERM_WINDOW(vwmterm),window,0,0,0,0,y-1,x-1,FALSE);

		/*	draw the cursor if the cursor_invisble flag is not set	*/
		vscreen=VWMTERM_SCREEN(vwmterm);
      if(!(vscreen->state & VSCREEN_CINVIS))
      {
         getyx(VWMTERM_WINDOW(vwmterm),y,x);
			mvwchgat(window,y,x,1,A_REVERSE,
				viper_color_pair(vscreen->fg,vscreen->bg),NULL);
      }

		/*	obligatory redraw and unlock	*/
		viper_window_redraw(window);
		viper_thread_leave();
	}

   /* if poll() returns -1, something bad has happened.  most likely the shell
      died (for some reason, maybe user sent SIGKILL).  if this happens we
      we need to clean up gracefully   */

	if(retval==-1 && errno!=EINTR)
   {
      viper_thread_enter();
      viper_window_close(window);
      viper_thread_leave();
   	/* allow this pseudo thread to run one more time so that waitpid() will
         catch the exit on next iteration.   */
     return PSTHREAD_CONTINUE;
   }

	if(retval>0) return PSTHREAD_GROW;
	return PSTHREAD_CONTINUE;
}


void vwmterm_carryover(vwmterm_t *vwmterm,gchar *pos_ptr,gint remainder)
{
   vwmterm->carryover=(gchar*)g_malloc0(remainder);
   memcpy(vwmterm->carryover,pos_ptr,remainder);
   vwmterm->carry_sz=remainder;

   return;
}

